/*
	AccountPolicy:

		Queries and sets User Account Policy for Windows NT from the command line.

		(Almost) functionally equivalent to the Policies->Account windows within
		User Manager.

		The only thing lacking that is in the windows is "Users must log on to change password."


		Copyright (C) Darren Tucker (1997,1998) (dtucker@zip.com.au)

		This program is free software; you can redistribute it and/or
		modify it under the terms of the GNU General Public License
		as published by the Free Software Foundation; either version 2
		of the License, or (at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

 /*
	Version History

	1997 11 15	Version 0.1
		Initial version.

	1998 08 16	Version 0.2
		Cleaned up code.
		Added -l (show license) option. 
		Added default multipliers.

  */

/*
	To rebuild you need to link with netapi32.lib

	To rebuild using MSVC use:
	cl accountpolicy.c /link netapi32.lib
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <lm.h>

static char PRGNAME[]="AccountPolicy";
static char PRGVER[]="v0.2";

#define	PWHISTORY		0
#define PWMAXAGE		1
#define	LOCKAFTER		2
#define	LOCKWINDOW		3
#define	LOCKDURATION	4
#define	PWMINAGE		5
#define	PWMINLENGTH		6

static char paramdesc[7][32]={"Password History", "Password Maximum Age", "Lock After",
		"Lock Window", "Lock Duration", "Password Minimum Age", "Password Minimum Length"};

/*	converts a string to a number of seconds

	accepts:
		char* to string to be converted.
		int parameter begin converted.

	returns: unsigned long number of seconds or
		TIMEQ_FOREVER if "Forever" or
		CONV_FAIL if the conversion failed.
 */
#define CONV_FAIL ((TIMEQ_FOREVER)-1)
unsigned long
convert_sec(char *str, int param)
{
	char *stopped, units, defunits='\0';
	unsigned long sec;

	if (strcmp(str,"0") == 0) {
		return(0);
	}
	
	_strlwr(str);
	if (strncmp(str, "fore", 4) == 0) {
		if (!(param == PWMAXAGE) || (param == LOCKDURATION)) {
			fprintf(stderr,"%s: Error: \"Forever\" inappropriate for %s\n", PRGNAME, paramdesc[param]);
			return(CONV_FAIL);
		}
		return(TIMEQ_FOREVER);
	}

	if ((sec=strtol(str, &stopped, 10)) == 0) {
		return(CONV_FAIL);
	}
	
	units = tolower(*stopped);	

	if (param==LOCKWINDOW || param==LOCKDURATION)
		defunits = 'm';
	if (param==PWMINAGE || param==PWMAXAGE)
		defunits = 'd';

	if (units == '\0') {		/* no units specified, use default */
		units = defunits;
	} else if ( units != defunits ) {
		fprintf(stderr, "%s: Warning: Units (%c) may be inappropriate for %s\n", PRGNAME, units, paramdesc[param]);
	}
   
	switch (units) { /* apply all units at and below supplied multiplier */
		case 'w': sec*=7;
		case 'd': sec*=24;
		case 'h': sec*=60;
		case 'm': sec*=60;
		case 's':
		case '\0': 
			break;
		default:
			return(CONV_FAIL);
	}

	if (sec < 0) {
		return(CONV_FAIL);
	} else {
		return(sec);
	}
}

/* print_usage:
	Um, well, prints usage info.
 */
void
print_usage()
{
	printf("%s %s: View/Set user account policy for the local machine.\n\n", PRGNAME, PRGVER);
	printf("usage:\t%s [-h] [-l] [PwMaxAge n] [PwMinAge n] [PwMinLen n]\n", PRGNAME);
	printf(		"\t[PwHistory n] [LockAfter n] [LockWindow n] [LockDuration n]\n\n");

	printf("Without parameters, it displays the current user account policy.\n\n");

	printf("Values of n are set using the appropriate units (ie the ones used in UserAdmin)\n");
	printf("however they can be specified by appending [smhdw] to the value, (eg 3m=3 mins)\n");
	printf("The keyword 'forever' can also be used for PwMaxAge and LockDuration.\n\n");

	printf("%s copyright (C) 1997,1998 Darren Tucker (dtucker@zip.com.au).\n", PRGNAME);
	printf("This is free software and comes with ABSOLUTELY NO WARRANTY.\n");
	printf("Run %s -l to show licence and copyright information.\n", PRGNAME);	
}

/* print_license:
	print license information
 */
void
print_license()
{
	printf("%s",

"AccountPolicy is copyright (C) 1997,1998 Darren Tucker (dtucker@zip.com.au)\n\
\n\
This program is free software; you can redistribute it and/or\n\
modify it under the terms of the GNU General Public License\n\
as published by the Free Software Foundation; either version 2\n\
of the License, or (at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program; if not, write to the Free Software\n\
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n"
	);
}

/* param_error:
	prints an error with the params and exits
 */
void
param_error(char *param, char *value)
{
	fprintf(stderr, "%s: Invalid value \"%s\" for %s.\n", PRGNAME, value, param);
	exit(1);
}

int
main (int argc, char **argv)
{
	PUSER_MODALS_INFO_0	modal0;
	PUSER_MODALS_INFO_3 modal3;
	LPWSTR	server=L"";
	int i;
	unsigned long sec;
	char *param;


	/* check for command line options */
	for(i=1; i<argc; i++){
		param = strdup(argv[i]);
		_strlwr(param);
		if (strncmp(param,"-",1)==0 || strncmp(param,"/",1)==0){
			if (strcmp(param, "-h")==0 || strcmp(param, "/h")==0 || strcmp(param, "-?")==0 || strcmp(param, "/?")==0 ) {
				print_usage();
				exit(0);
			} else if (strcmp(param, "-l")==0 || strcmp(param, "/l")==0) {
				print_license();
				exit(0);
			} else {
				fprintf(stderr, "%s: Error: Unknow option %s.\n", PRGNAME, argv[i]);
				exit(1);
			}
		}
		free(param);
	}

	/* Allocate policy buffers and get current policy */
	if (NetApiBufferAllocate(sizeof(USER_MODALS_INFO_0), &modal0) != NERR_Success ||
		NetApiBufferAllocate(sizeof(USER_MODALS_INFO_3), &modal3) != NERR_Success) {
		fprintf(stderr, "%s: could not allocate NetApiBuffer.\n", PRGNAME);
		exit(1);
	}

	if (NetUserModalsGet(server, 0, (LPBYTE *)&modal0) != NERR_Success ||
		NetUserModalsGet(server, 3, (LPBYTE *)&modal3) != NERR_Success) {
		fprintf(stderr, "%s: could not get current policy.", PRGNAME);
		exit(1);
	}

	/* parse remaining command line and fill in any changes to policy */
	/* NOTE: order is important for processing options. Shortest unique matches must be first.
		Current valid values are and the number of character to make the option unique:
			PwHistory		3
			PwMaxAge		4
			LockAfter		5
			LockWindow		5
			LockDuration		5
			PwMinAge		6
			PwMinLength		6
	 */

	for(i=1; i<argc; i++){
		param = strdup(argv[i]);
		_strlwr(param);
		if (strncmp(argv[i], "pwhistory", 3) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],PWHISTORY)) != CONV_FAIL) {
				modal0->usrmod0_password_hist_len = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "pwmaxage", 4) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],PWMAXAGE)) != CONV_FAIL) {
				modal0->usrmod0_max_passwd_age = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "lockafter", 5) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],LOCKAFTER)) != CONV_FAIL) {
				modal3->usrmod3_lockout_threshold = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "lockwindow", 5) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],LOCKWINDOW)) != CONV_FAIL) {
				modal3->usrmod3_lockout_observation_window = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "lockduration", 5) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],LOCKDURATION)) != CONV_FAIL) {
				modal3->usrmod3_lockout_duration = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "pwminage", 6) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],PWMINAGE)) != CONV_FAIL) {
				modal0->usrmod0_min_passwd_age = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else if (strncmp(argv[i], "pwminlength", 6) == 0) {
			if (++i < argc && (sec=convert_sec(argv[i],PWMINLENGTH)) != CONV_FAIL) {
				modal0->usrmod0_min_passwd_len = (DWORD)sec;
			} else {
				param_error(argv[i-1], argv[i]);
			}
		} else {
			fprintf(stderr, "%s: parameter not recognised: %s\n", PRGNAME, argv[i]);
		}

		free(param);
	}

	if (argc == 1) {
		printf("Current user account policy on local machine:\n\n");
		printf("PwMaxAge\tMaximum password age: ");
		if (modal0->usrmod0_max_passwd_age == TIMEQ_FOREVER) {
			printf("forever.\n");	
		} else {
			printf("%lu days.\n", modal0->usrmod0_max_passwd_age/24/60/60);
		}
		printf("PwMinAge\tMinimum password age: %lu days.\n", modal0->usrmod0_min_passwd_age/60/60/24);
		printf("PwMinLen\tMinimum password length: %lu characters.\n", modal0->usrmod0_min_passwd_len);
		printf("PwHistory\tRemember %lu unique passwords.\n\n", modal0->usrmod0_password_hist_len);

		printf("LockAfter\tLockout account after %lu logon attempts.\n", modal3->usrmod3_lockout_threshold);
		printf("LockWindow\tReset lockout count after %lu minutes.\n", modal3->usrmod3_lockout_observation_window/60);
		printf("LockDuration\tLockout account ");
		if (modal3->usrmod3_lockout_duration == TIMEQ_FOREVER) {
			printf("forever.\n");	
		} else {
			printf("for %lu minutes.\n", modal3->usrmod3_lockout_duration/60);
		}
		
		exit(0);
	}

	if (NetUserModalsSet(server, 0, (LPBYTE)modal0, NULL) != NERR_Success ||
		NetUserModalsSet(server, 3, (LPBYTE)modal3, NULL) != NERR_Success) {
		fprintf(stderr, "%s: Error: could not set new policy.", PRGNAME);
		exit(1);
	}

	NetApiBufferFree(modal0);
	NetApiBufferFree(modal3);

	return(0);
}
